#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "ts_msgs.h"
#include "comp.h"

/* initialize the seed */
#define init_seed  ( srand ( clock () + ( unsigned int ) time ( NULL ) ) )

/* generate a random integer number between zero and a-1, inclusively */
#define mrandom(a) ( rand () % (a) )

/* generate a random double number between zero and one, inclusively */
#define rnd        ( ( double ) rand () ) / ( ( double ) RAND_MAX )

/* these variables describe the resulting triangle strip representation
 *
 *     ds        the number of vertices of the current triangle strip
 *     dds       the sum of the lengths of triangle strips in the representation
 *     ps        the number of triangle strips in the representation
 *     minl      the length of the shortest triangle strip
 *     maxl      the length of the longest triangle strip 
 *     cnt[]     if i < 9, cnt[i] is the number of triangle strips of length i + 1 in the representation;
 *               cnt[9] is the number of triangle strips in the representation the length of which is at least 10 */
long ds, dds, ps, minl, maxl;
long cnt[10];

/* procedure inithn initializes the Hopfield network
 */
void inithn ( void ) 
{
	long cn, j;
	
	cn = 2 * es_size;
	for ( j = 0; j < hn_size; j++ ) {
		if ( j < cn )
			/* randomly generate the initial state of a neuron that is
			 * assigned to an edge */
			hn[j].s = (char) mrandom ( 2 );
		else
			/* the initial states of d_C and a_C neurons are zero */
			hn[j].s = (char) 0;

		/* set the has-been-handled indicator to zero for any neuron */
		hn[j].u = (char) 0;
	}
}

/* procedure prints prints the actual state (that is, the state at macroscopic time t)
 * of the Hopfield network into the file out
 */
void prints ( long t, FILE *out )
{
	int j;
	
	fprintf ( out, "\nThe state of Hopfield net at time %u:\n", t );
	
	for ( j = 0; j < hn_size; j++ ) {
		fprintf ( out, "%d", (int) hn[j].s );
		if ( j < hn_size - 1 && j % 60 == 59 ) fprintf ( out, "\n" );
	}
	
	fprintf ( out, "\n" );
}

/* procedure printv prints the list of all vertices into the file rout
 */
void printv ( FILE *rout )
{
	int i;
	
	for ( i = 0; i < vs_size; i++ )
		fprintf ( rout, "v %g %g %g\n", vs[i].x, vs[i].y, vs[i].z );
}

/* function comval returns the common vertex of edges e1, e2
 */
long comval ( long e1, long e2 )
{
	long a1 = es[e1].v1;
	long a2 = es[e1].v2;
	long b1 = es[e2].v1;
	long b2 = es[e2].v2;
	
	if ( a1==b1 || a1==b2 ) return a1;
	if ( a2==b1 || a2==b2 ) return a2;

	/* report abnormal termination */
        ts_warning ( "comval", 5 );
	return -1;
}

/* procedure strip writes a vertex v of a triangle strip into
 * the file rout
 */
void strip ( long v, FILE *rout )
{
	/* write "ts " at the beginning of a strip */
	if ( ++ds == 1 ) fprintf ( rout, "t " );
	
	/* otherwise, write "q " */
	else  fprintf ( rout, "q " );

	/* write the number of the vertex (increased by one since 
	 * vertices are indexed starting from zero in the array) */
	fprintf ( rout, "%u\n", v + 1 );
}

/* procedure estrip writes a message into the file rout
 * at the end of a strip, and updates related variables
 */
void estrip ( FILE *rout )
{
	fprintf ( rout, "# tristrip length: %u\n", ds - 2 );
	
	/* increase the number of strips */
	ps++;

	/* adjust the minimal and the maximal length if necessary */
	if ( ds - 2 < minl ) minl = ds - 2;
	if ( ds - 2 > maxl ) maxl = ds - 2;

	/* adjust the counts of strip lengths */
	if ( ds - 2 < 10 ) 
		cnt[ ds - 3 ]++;
	else
		cnt[ 9 ]++;
	
	/* increase the sum of lengths of all strips in the representation */
	dds += ds - 2;
}

/* function prob calculates the probability of activation according
 * to the formula 
 *                        1
 *        prob = -------------------
 *               1 + exp ( -2*s/tp )
 */
double prob ( long s, double tp )
{
	double x;
	
	x = (double) (2 * s) / tp;

	/* if x is small enough, return the exact value */
	if ( fabs ( x ) < 20 ) return 1 / ( 1 + exp ( -x ) );
	
	/* otherwise, round the value appropriately */
	if ( x >= 0 ) return 1;
	return 0;
}

/* procedure comp performs the computation by Hopfield network and writes
 * the details about this performance into the file "comp.txt"
 *
 *     fname     the name of the input .obj file
 *     p_tt      a pointer to the variable that contains the convergence
 *               macroscopic time of the computation at the output
 *     tp0       initial temperature for simulated annealing
 *     end       stopping criterion (simulated annealing will stop if the
 *               ratio of the neurons that changed the state during current
 *               iteration is less than or equal to end)
 */
void comp ( char *fname, long *p_tt, double tp0, double end )
{
	int ch, k;
	long cn, i, j, n, nu, ns, pch, s, t;
	double tp;
	const double ln2 = log ( (double) 2.0 );
	FILE *out;

	out = fopen ( "tstr_tmp.txt", "w" );
	if ( !out ) ts_error ( "comp" , 2 );
	fprintf ( out, "Computation by Hopfield net for \"%s\"\n", fname );
	
	/* initialize the seed */
	init_seed;

	/* initialize the Hopfield network */
	inithn ();
	
	/* print the state of the Hopfield network at macroscopic time 0 */
	prints ( 0, out );
	
	/* set the macroscopic time */
	t = 1;

	/* cn will be the number of neurons in the network */
	cn = hn_size;

	/* set the temperature */
	tp = tp0;

	/* convert the stopping criterion ratio to a rational number */
	end = end / 100.0;
	
	/* perform simulated annealing (k=0), then compute with 
	 * the use of the Heaviside activation function (k=1) */
	for ( k = 0; k < 2; k++ )

		/* repeat until the ratio of the neurons that changed the state
		 * is small enough (k=0) or until there is no neuron that changed
		 * its state (k=1) */
		do {

			/* pch counts the neurons that changed their states */
			pch = 0;

			/* work around all the neurons */
			for ( nu = 0; nu < cn; nu++ ) {

				/* choose a neuron randomly */
				j = mrandom ( cn );
				
				/* if the neuron has already been handled at this macroscopic time,
				 * shift to its nearest successor in hn than is still to be handled;
				 * if you hit the end of the array, continue at the beginning */
				while ( (int) hn[j].u == t % 2 ) j = ( j + 1 ) % cn;
				
				/* mark the neuron as handled at this macroscopic time */
				hn[j].u = (char) ( t % 2 );

				/* the excitation of the neuron will be computed to s */
				s = hn[j].w0;
				
				n = hn[j].c_size;

				/* for d_C, do not consider j_C to be a neighbour, i.e. compute with w(j_C,d_C)=0 */
				if ( j % 2 == 0 && j / 2 >= es_size ) n--; 	
				
				/* compute the excitation of the neuron */
				for ( i = 0; i < n; i++ )
					s += hn[j].w[i] * (int) hn[ hn[j].c[i] ].s;
				
				/* compute the new state of the neuron */
				if ( k ) ns = ( s >= 0 ); else ns = ( rnd <= prob ( s, tp ) );
				
				/* increase pch if the state of the neuron is going to be changed */ 
				pch += ( (int) hn[j].s != ns );

				/* update the state of the neuron */
				hn[j].s = (char) ns;
			}
			
			/* if ch, a next iteration (with k unchanged) is needed */
			if ( k ) ch = ( pch > 0 ); else ch = ( (double) pch / (double) cn > end );
			
			/* update the temperature if needed */
			if ( !k && ch ) tp = ln2 * tp0 / log ( (double) (t + 1) );

			/* if a next iteration is going to be performed, print the state of
			 * the network and increase the macroscopic time */
			if ( !k || ch ) prints ( t++, out );
		}
		while ( ch );

	/* return the convergence macroscopic time */
	*p_tt = t - 1;
	
	if ( fclose ( out ) == EOF ) ts_warning ( "comp", 4 );
}

/* function trisgen generates an .objf representation
 *
 *     fname     the name of the input .obj file
 *     ofname    the name of the file in which the representation should be written
 *     fps       a pointer to the variable the output value of which is the number
 *               of triangle strips in the representation
 *     fminl     a pointer to the variable the output value of which is the length
 *               of the shortest triangle strip in the representation
 *     fmaxl     a pointer to the variable the output value of which is the length
 *               of the longest triangle strip in the representation
 */
int trisgen ( char *fname, char *ofname, long *fps, long *fminl, long *fmaxl )
{ 
	int b, k, u, db;
	long i, j, pe, pee, pj, pjj, v;
	FILE *rout;

	/* initialize the total of the lengths of triangle strips, the length of the shortest and of the longest strip,
	 * and the array of counts of strip lengths */
	dds = 0;
	minl = ntr + 1;
	maxl = -1;
	for ( j = 0; j < 10; j++ ) cnt[j] = 0;

	rout = fopen ( ofname, "w" );
	if ( !rout ) ts_error ( "trisgen", 2 ); 
	fprintf ( rout, "# sequential triangle strip representation for \"%s\"\n", fname );
	fprintf ( rout, "# by Jiri Sima and Radim Lnenicka (lnenicka at utia.cas.cz)\n" );
	
	/* print the list of vertices */
	printv ( rout );
	
	/* initialize the number of triangle strips */
	ps = 0;

	/* set the has-been-handled flag */
	u = 1 - (int) hn[0].u;
	
	/* do for all neurons that are associated with an edge */
	for ( j = 0; j < 2 * es_size; j++ ) {

		/* test whether the state of the neuron is 1 and whether the neuron is still to be handled */
		if ( (int) hn[j].s && (int) hn[j].u != u ) {

			/* get the edge associated with the neuron */
			pe = j / 2;

			/* get the indices of the neighbours of pe that correspond to the strip controlled by 
			 * the neuron */
			pj = ( j + (int) es[pe].r ) % 2;
			pjj = ( pj + 2 ) % 4;
			
			/* handle the neuron only if it corresponds to an end of a strip; in this case,
			 * let pj point towards the whole strip, not to its end */
			if ( ( pee = es[pe].e[pjj] ) >= 0 && 
					(int) hn[  2 * pee + ( (int) es[pe].ie[pjj] + (int) es[pee].r ) % 2  ].s )
				if ( ( pee = es[pe].e[pj] ) < 0 ||
						!( (int) hn[  2 * pee + ( (int) es[pe].ie[pj] + (int) es[pee].r ) % 2  ].s ) )
					pj = pjj;
				else
					continue;
			
			/* reset the counter of vertices in the strip */
			ds = 0;
			
			/* try to find the index db of a neighbour of pe that is an internal edge */
			for ( db = 0; db < 4; db++ )
				if ( es[pe].e[db] >= 0 ) break;

			if ( db < 4 ) {

				/* if there is such a neighbour db, let v be the common vertex of db and pe */
				v = comval ( es[pe].e[db], pe );
				
				/* the cases
				 *
				 *            *                               *
				 *          /   \ e[pj] = e[db]             /   \ e[pj]
				 *        /  pe   \                       /  pe   \
				 *    b * --------- * c = v   and     b * --------- * c = v
				 *        \       /                       \       /
				 *          \   / pee                       \   / pee = e[db]
				 *          a *                             a *                               */
				if ( db == pj || db == 3 - pj ) {

					/* print the vertex a first */
					if ( ( pee = es[pe].e[ 3 - pj ] ) < 0 ) strip ( -pee - 1, rout );
					else if ( v == es[pee].v1 ) strip ( es[pee].v2, rout );
					else strip ( es[pee].v1, rout );
					
					/* print the vertex b */
					if ( v == es[pe].v1 ) strip ( es[pe].v2, rout );
					else strip ( es[pe].v1, rout );
					
					/* print the vertex c */
					strip ( v, rout );
				}
				/* the cases
				 *
				 *             *                               *
				 *     e[db] /   \ e[pj]                     /   \ e[pj]
				 *         /  pe   \                       /  pe   \
				 * b = v * --------- * c   and     b = v * --------- * c
				 *         \       /                       \       /
				 *      pee  \   /               pee = e[db] \   /
				 *           a *                             a *                               */
				else {
					/* print the vertex a first */
					if ( ( pee = es[pe].e[  ( pj + 2 ) % 4  ] ) < 0 ) strip ( -pee - 1, rout );
					else if ( v == es[pee].v1 ) strip ( es[pee].v2, rout );
					else strip ( es[pee].v1, rout );
					
					/* print the vertex b */
					strip ( v, rout );
					
					/* let v be the vertex c for further use */
					if ( v == es[pe].v1 ) v = es[pe].v2;
					else v = es[pe].v1;

					/* print the vertex c */
					strip ( v, rout );
				}
			}
			else {
				/* when none of the neighbours of pe are internal edges, the vertices 
				 * can be printed in either of the orders a, b, c or a, c, b */
				
				pee = es[pe].e[ 3 - pj ];
				
				/* print the vertex a first */
				strip ( -pee - 1, rout );

				/* print the vertices b, c */
				strip ( es[pe].v1, rout );
				strip ( es[pe].v2, rout );
			}
			
			/* get the index in hn of the current neuron */
			i = 2 * pe + ( pj + (int) es[pe].r ) % 2;
				
			do {
				/* set the has-been-handled attribute of the neuron i */
				hn[i].u = (char) u;
				
				/* b=1 if the next edge along the strip is internal,
				 * b=0 otherwise */
				b = ( ( pee = es[pe].e[pj] ) >= 0 );
				
				if ( b ) {
					/* if the next edge is internal, update pj */
					pj = ( (int) es[pe].ie[pj] + 2 ) % 4;
						
					/* update v */
					if ( v == es[pee].v1 ) v = es[pee].v2;
					else v = es[pee].v1;
					
					/* print the next vertex of the strip */
					strip ( v, rout );

					/* update pe */
					pe = pee;
				}
				else
					/* if the next edge is boundary, print the terminating
					 * vertex of the strip */
					strip ( -pee - 1, rout );
			}
			/* repeat while the next edge along the strip is internal and the state of the next
			 * neuron along the strip is 1 */
			while ( b && (int) hn[  i = 2 * pe + ( pj + (int) es[pe].r ) % 2  ].s );
			
			/* write a message at the end of the strip and update counters */
			estrip ( rout );
		}
	}
	
	/* record internal triangle strips of length 1, i.e. triangles that alone form strips but that are not 
	 * isolated; these strips are searched for by going over all internal edges such that their associated 
	 * neurons are both in zero states */
	for ( i = 0; i < es_size; i++ )

		/* consider such an internal edge i only if it has not been handled already */
		if ( !( (int) hn[ 2 * i ].s ) && !( (int) hn[ 2 * i + 1 ].s ) && (int) hn[ 2 * i ].u != u ) {
			
			/* set the boolean indicating that a triangle which the edge i is part of 
			 * represents a triangle strip that should be printed */
			b = 1;

			/* take the neighbours of i successively; the two neighbours of i with k=0 and k=1
			 * and i form one triangle, the two with k=2 and k=3 and i another one */
			for ( k = 0; k < 4; k++ ) {
				
				/* reset the boolean if the triangle is a part of a triangle strip the length of
				 * which is greater than 1 or if it has been handled already via its another edge */
				b = b && ( ( pe = es[i].e[k] ) < 0 || 
					( (int) hn[ 2 * pe ].s == 0 && (int) hn[ 2 * pe + 1 ].s == 0 &&
					  (int) hn[ 2 * pe ].u != u ) );
				
				if ( k % 2 == 1 )
					/* all the testing of a triangle has been done */
					if ( b ) {
						/* print the triangle strip */

						/* reset the counter of vertices */
						ds = 0;
						
						/* print the vertices in any order */
						if ( pe < 0 ) strip ( -pe - 1, rout );
						else if ( comval ( pe, i ) == es[pe].v1 ) strip ( es[pe].v2, rout );
						else strip ( es[pe].v1, rout );
						
						strip ( es[i].v1, rout );
						strip ( es[i].v2, rout );
						
						/* end the strip */
						estrip ( rout );
					}
					/* set the boolean to 1 for testing the other triangle */
					else b = 1;
			}
			
			/* mark the edge as handled */
			hn[ 2 * i ].u = (char) u;
		}

	/* record triangle strips arising from isolated triangles */
	for ( i = 0; i < aes_size; i += 3 )
		
		if ( aes[i].ies == -2 ) {
			
			/* reset the counter of vertices */
			ds = 0;
			
			/* write the vertices */
			strip ( aes[i].v1, rout );
			strip ( aes[i].v2, rout );
			strip ( aes[i + 1].v2, rout );

			/* end the strip */
			estrip ( rout );
		}
		
	fprintf ( rout, "#\n" );
	
	/* the total of the lengths of the triangle strips in the representation should equal
	 * the total number of triangles read from the input .obj file */
	if ( dds == ntr ) {
		fprintf ( rout, "# number of sequential tristrips: %u\n", ps );
		
		/* return the number of strips, the length of the shortest and of the longest strip */
		*fps = ps;
		*fminl = minl;
		*fmaxl = maxl;
			
		if ( fclose ( rout ) == EOF ) ts_warning ( "trisgen", 4 );
		return 0;
	}
	else {
		/* this occurs only if the surface given by the input .obj file violates some of the assumptions
		 * given to it or if there is a bug somewhere in the code */
		fprintf ( rout, "# ERROR: %u triangles missing\n", labs ( ntr - dds ) );
		if ( fclose ( rout ) == EOF ) ts_warning ( "trisgen", 4 );
		return 1;
	}
}
