#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ts_msgs.h"
#include "ts_types.h"
#include "input.h"

/* initialize the number of triangles read from the input file */
long ntr = 0;

/* initialize the current line of the input file */
long line = 0;

/* a pointer to the array of vertices */
vertex *vs;

/* a pointer to the array of all sides (edges) of all triangles */
aedge *aes;

/* a pointer to the array of all internal edges */
edge *es;

/* the lengths of the corresponding arrays */
long vs_size, aes_size, es_size;

/* function compare defines comparison of long integers that represent indices
 * of aedges in aes; 
 *
 *    index i is greater than index j if aes[i].v1 > aes[j].v1, 
 *            
 *                              or if aes[i].v1 = aes[j].v1 and
 *                                    aes[i].v2 > aes[j].v2;
 *
 *    they are equal if aes[i].v1 = aes[j].v1 and aes[i].v2 = aes[j].v2;
 *
 *    otherwise, index j is greater than index i 
 */
int compare ( const void *a, const void *b )
{
	long temp1, temp2;
	
	temp1 = ( aes + *((long *) a) ) -> v1 - ( aes + *((long *) b) ) -> v1;
	temp2 = ( aes + *((long *) a) ) -> v2 - ( aes + *((long *) b) ) -> v2;
	
	if ( temp1 == 0 ) {
		if ( temp2 > 0 )
			return 1;
		else if ( temp2 < 0 )
			return -1;
		else
			return 0;
	}
	else if ( temp1 > 0 )
		return 1;
	else
		return -1;
}

/* function compare_ord defines ordinary numerical comparison of long int numbers */
int compare_ord ( const void *a, const void *b )
{
	long temp;

	temp = *((long *) a) - *((long *) b);
	if ( temp > 0 )
		return 1;
	else if ( temp < 0 )
		return -1;
	else
		return 0;
}

/* Procedure getln reads a line from the file file and stores it in a buffer that is
 * addressed by the pointer (*lineptr). No end-line character is stored at the end
 * of the line in the buffer but a single terminating '\0' character is added.
 *
 * If a backslash character is hit during reading the line from the file, it is replaced 
 * by a space character, the rest of the current line is ignored, and the next line of
 * the file is appended to the buffer.
 *
 * The procedure supposes that a memory space of length (*size) characters is allocated 
 * for the line beforehand and reallocates more memory if needed. If there is such
 * a need, the memory is reallocated with addition of space for more INIT_BUFFER_SIZE 
 * characters. The INIT_BUFFER_SIZE macro is defined in the 'header.h' file.
 * The length of the (re-)allocated memory is stored in the variable addressed by the
 * pointer size. 
 * 
 * For portability reasons, any sequence of LF and CR characters is considered as the end 
 * of a line; as a result, blank lines of the file are ignored except for the first
 * line of the file.
 */
void getln ( char **lineptr, long *size, FILE *file )
{
	long len;
	unsigned char c, c_skip;
	
	/* increase the current line number */ 
	line++;

	/* set the length of the line stored in the buffer to zero */
	len = 0;
	
	/* repeat until the end of the line or the end of the file is reached */ 
	do {
		c = ( unsigned char ) getc ( file );
		
		if ( !feof ( file ) ) {
			
			/* handle the backslash */
			if ( c == '\\' ) {

				/* prepare the replacement by a space */
				c = ' ';

				/* ignore the rest of the line */
				do {
					c_skip = ( unsigned char ) getc ( file );
				}
				while ( !feof ( file ) && c_skip != '\xd' && c_skip != '\xa' );

				do {
					c_skip = ( unsigned char ) getc ( file );
				}
				while ( c_skip == '\xd' || c_skip == '\xa' );

				if ( !feof ( file ) ) ungetc ( c_skip, file );
			}
			
			/* do not store '\xd' and '\xa' in the buffer */
			if ( c != '\xd' && c != '\xa' ) {

				/* reallocate memory if needed for storing a character */
				if ( len == *size ) {
					*size += INIT_BUFFER_SIZE;
					*lineptr = ( char * ) realloc ( ( void * ) *lineptr, ( *size ) * sizeof ( char ) );
					if ( !( *lineptr ) ) ts_error ( "getln", 3 );
				}

				/* store the character */
				*( *lineptr + len ) = c;

				/* increase the length of the line */
				len++;
			}
		}
	}
	while ( !feof ( file ) && c != '\xd' && c != '\xa' );

	/* if the line terminates by '\xd' or '\xa', skip the whole end-of-line sequence */
	if ( !feof ( file ) ) {
		do {
			c = ( unsigned char ) getc ( file );
		}
		while ( c == '\xd' || c == '\xa' );

		if ( !feof ( file ) ) ungetc ( c, file );
	}
	
	/* reallocate memory if needed for storing the terminating '\0' */
	if ( len == *size ) {
		*size += INIT_BUFFER_SIZE;
		*lineptr = ( char * ) realloc ( ( void * ) *lineptr, ( *size ) * sizeof ( char ) );
		if ( !( *lineptr ) ) ts_error ( "getln", 3 );
	}

	/* store '\0' */
	*( *lineptr + len ) = '\0';
}

/* Procedure wrongformat displays an error message and exits the program.
 * This occurs when a (logical) line of the input file that starts with 
 * 'v' (vertex) or with 'f' (face) does not contain three more tokens.
 */
void wrongformat ( void )
{
	printf ( "\nERROR!!! rfile: Wrong format of line %u of input file.\n", line );
	exit ( -1 );
}

/* Procedure rfile parses the lines of the input file fname and creates the arrays
 * vs of vertices and aes of all sides (edges) of all triangles. For each 
 * triangle, its edges are stored in aes in the order xy, xz, yz, where x < y < z
 * are the vertices of the triangle.
 */
void rfile ( const char *fname ) 
{
	FILE *file;
	char *buffer, **p_buffer, *token, *p, *x, *y, *z;
  	long b_size, aes_len, vs_len, v[3];  
  	int i;
	const char delimiters[] = " \t";

	/* set the buffer size */
	b_size = INIT_BUFFER_SIZE;

	/* allocate the memory for the buffer */
	buffer = ( char * ) malloc ( b_size * sizeof ( char ) );
	if ( !buffer ) ts_error ( "rfile", 3 );
	
	p_buffer = &buffer;

	file = fopen ( fname, "r" );
	if ( !file ) ts_error ( "rfile", 2 );
	clearerr ( file );
	
	/* initialize the lengths of the vs and aes arrays */
	vs_size = aes_size = 0;
 
	/* allocate memory for initial amounts of elements of the arrays */
	vs_len = aes_len = LENGTH_OF_BLOCK;
	vs = ( vertex * ) malloc ( vs_len * sizeof ( vertex ) );
	aes = ( aedge * ) malloc ( aes_len * sizeof ( aedge ) );
	if ( !vs || !aes ) ts_error ( "rfile", 3 );
	
	while ( !feof ( file ) ) {
		
		/* read a line from the file fname */
		getln ( p_buffer, &b_size, file );

		/* get the first token of the line */
		token = strtok ( *p_buffer, delimiters );
		
		/* if the token is 'v', parse a vertex */
		if ( token && !strcmp ( token, "v" ) ) {

			/* reallocate memory if needed */
			if ( vs_size == vs_len ) {
				vs_len += LENGTH_OF_BLOCK;
				vs = ( vertex * ) realloc ( ( void * ) vs, vs_len * sizeof ( vertex ) );
				if ( !vs ) ts_error ( "rfile", 3 );
			}
			
			/* get the coordinates of the vertex as strings */
			x = strtok ( NULL, delimiters );
			y = strtok ( NULL, delimiters );
			z = strtok ( NULL, delimiters );

			/* exit if there is not enough tokens in the line */
			if ( !x || !y || !z ) wrongformat ();
			
			/* convert the coordinates to doubles and save them */
			vs[ vs_size ].x = strtod ( x, NULL );
			vs[ vs_size ].y = strtod ( y, NULL );
			vs[ vs_size ].z = strtod ( z, NULL );
			
			/* increase the number of vertices */
			vs_size++;
		}
		/* if the token is 'f', parse a face */
		else if ( token && !strcmp ( token, "f" ) ) {
			
			/* increase the number of triangles */
			ntr++;
			
			for ( i = 0; i < 3; i++ ) {
				
				/* get the information about the (i+1)-th vertex of the face */
				token = strtok ( NULL, delimiters );
				
				/* exit if there is no more token in the line */
				if ( !token ) wrongformat ();
				
				/* if there is a slash in the token, ignore it as well as the rest of the token */
				p = strchr ( token, '/' );
				if ( p ) *p = '\0';
				
				/* get the number corresponding to the (i+1)-th vertex of the face */
				v[i] = strtol ( token, NULL, 10 );
				if ( v[i] < 0 ) 
					/* if v[i] is negative, then it refers to (-v[i])-th last read vertex;
					 * recalculate the identification of the vertex */
					v[i] += vs_size;
				else
					/* otherwise only decrease the number by one since the vertices are indexed
					 * starting from zero */
					v[i]--;
			}
			
			/* sort the vertices in ascending order */
			qsort ( v, 3, sizeof ( long ), &compare_ord );
			
			/* if there is not enough allocated memory to store another three edges, reallocate */
			if ( aes_size + 3 > aes_len ) {
				aes_len += LENGTH_OF_BLOCK;
				aes = ( aedge * ) realloc ( (void *) aes, aes_len * sizeof ( aedge ) );
				if ( !aes ) ts_error ( "rfile", 3 );
			}
			
			/* store the edges and increase aes_size accordingly */
			( aes + aes_size ) -> v1 = v[0];
			( aes + aes_size ) -> v2 = v[1];
			aes_size++;

			( aes + aes_size ) -> v1 = v[0];
			( aes + aes_size ) -> v2 = v[2];
			aes_size++;

			( aes + aes_size ) -> v1 = v[1];
			( aes + aes_size ) -> v2 = v[2];
			aes_size++;
		}
	}

	/* free the buffer and close the input file */
	free ( (void *) (*p_buffer) );
	if ( fclose ( file ) == EOF ) ts_warning ( "rfile", 4 );
}

/* Procedure nb gets the index i of an aedge aes[i], and sets the variables at addresses
 * e1 and e2 to contain the indices of the remaining two aedges that belong to the same 
 * triangle as aes[i].
 */
void nb( const long i, long *e1, long *e2) 
{
	/* m is the least index in aes that belongs to the triangle of aes[i] */
	long m = (i / 3) * 3;
	
	*e1 = m + ( (i + 1) % 3 );
	*e2 = m + ( (i + 2) % 3 );
}

/* procedure input reads the data from the input file fname and creates the array vs
 * of strings that represent vertices and the array es of all internal edges; it also
 * temporarily creates and uses the array aes of all sides (edges) of all triangles
 * (i.e. the array of aedges)
 */
void input ( char *fname )
{
	long *sa, i, cnt;
	int j, k;
	long e1, e2, pe, s1, s2, v;
  
	/* convert the file fname to the arrays vs and aes */
	rfile ( fname );
	
	/* initialize the array sa that will encode the ordering of aedges in aes */
	sa = ( long * ) malloc ( aes_size * sizeof ( long ) );
	if ( !sa ) ts_error ( "input", 3 );
	for ( i = 0; i < aes_size; i++ )
		sa[i] = i;
	
	/* sort the elements of sa, which represent the indices of aedges in aes,
	 * according to the comparison function compare */
	qsort ( sa, aes_size, sizeof ( long ), &compare );
	
	/* set the ies attribute of every aedge in aes; a given aedge may occur once
	 * or twice in aes, depending on whether it represents a side of only one triangle
	 * or a common side of two triangles, respectively; the ies attribute is set
	 * to -1 for all aedges that occur only once in aes; the ies attributes of both
	 * occurences of an aedge that occurs twice in aes are set to a single unique
	 * identification number */
	for ( i = 0, cnt = 0; i < aes_size; i++ ) {
		if ( ( i < aes_size - 1 ) && !compare ( (void *) (sa + i) , (void *) (sa + i + 1) ) ) {
			aes[ sa[i++] ].ies = cnt;
			aes[ sa[i]   ].ies = cnt++;  
		}
		else aes[ sa[i] ].ies = -1;
	}
	
	/* allocate memory for the array es of all internal edges */
	es = ( edge * ) malloc ( cnt * sizeof ( edge ) );
	if ( !es ) ts_error ( "input", 3 );
	
	/* initialize the length of es */
	es_size = 0;

	/* fill the array es with internal edges */
	for ( i = 0; i < aes_size; i++ )
		
		/* test whether aes[ sa[i] ] is an internal edge */ 
		if ( aes[ sa[i] ].ies != -1 ) {

			/* if so, copy the vertices of aes[ sa[i] ] to a newly created edge es[es_size] */
			es[es_size].v1 = aes[ sa[i] ].v1;
			es[es_size].v2 = aes[ sa[i] ].v2;
			
			/* find the remaining sides of the triangle of aes[ sa[i] ] */
			nb( sa[i], &e1, &e2 );

			/* declare e1 (and e2) to be a neighbour of es[es_size] if it is an internal edge,
			 * otherwise, set the corresponding attribute of es[es_size] to -1 */
			es[es_size].e[0] = aes[e1].ies;
			es[es_size].e[1] = aes[e2].ies;
			
			/* find the common vertex v of e1 and e2 */
			if ( ( aes[e1].v1 == aes[e2].v1 ) || ( aes[e1].v1 == aes[e2].v2) )
				v = aes[e1].v1;
			else 
				v = aes[e1].v2;
			
			/* if a neighbouring edge is boundary, let it point to v (using a transformed value of v) */ 
			v++;
			if ( es[es_size].e[0] == -1 ) es[es_size].e[0] = -v;
			if ( es[es_size].e[1] == -1 ) es[es_size].e[1] = -v;
			
			/* prepare auxiliary variables for fixing mutual orientation of neighbouring triangles */
			s1 = aes[e1].v1 + aes[e1].v2;
			s2 = aes[e2].v1 + aes[e2].v2;
			
			/* find the remaining edges of the triangle of aes[ sa[++i] ], i.e. of the triangle that
			 * shares the edge es[es_size] with the triangle considered so far */
			nb( sa[++i], &e1, &e2);

			/* fix mutual orientation of the two neighbouring triangles, so that
			 *   these configurations are allowed:            while these are not:
			 *            *                  *                         *                  *
			 *     e[0] /   \ e[1]    e[1] /   \ e[0]           e[0] /   \ e[1]    e[1] /   \ e[0]
			 *        /es_size\          /es_size\                 /es_size\          /es_size\
			 *      * --------- *      * --------- *             * --------- *      * --------- * 
			 *        \       /          \       /                 \       /          \       /
			 *     e[3] \   / e[2]    e[2] \   / e[3]           e[2] \   / e[3]    e[3] \   / e[2]
			 *            *                  *                         *                  *           */
			s1 += aes[e1].v1 + aes[e1].v2;
			s2 += aes[e2].v1 + aes[e2].v2;
			
			if ( s1 != s2 ) {
				pe = e1;
				e1 = e2;
				e2 = pe;
			}
			
			/* declare the new e1 (and e2) to be a neighbour of es[es_size] if it is an internal edge,
			 * otherwise, set the corresponding attribute of es[es_size] to -1 */
			es[es_size].e[2] = aes[e1].ies;
			es[es_size].e[3] = aes[e2].ies;
		
			/* find the common vertex v of e1 and e2 */
			if ( ( aes[e1].v1 == aes[e2].v1 ) || ( aes[e1].v1 == aes[e2].v2 ) )
				v = aes[e1].v1;
			else
				v = aes[e1].v2;
			
			/* if a neighbouring edge is boundary, let it point to v (using a transformed value of v) */
			v++;
			if ( es[es_size].e[2] == -1 ) es[es_size].e[2] = -v;
			if ( es[es_size].e[3] == -1 ) es[es_size].e[3] = -v;
			
			/* increase the variable that counts elements of es */
			es_size++;
		}

	/* mark the triangles that have only boundary edges -- these triangles are isolated */
	for ( i = 0; i < aes_size; i += 3 )
		if ( aes[i].ies == -1 && aes[i + 1].ies == -1 && aes[i + 2].ies == -1 ) aes[i].ies = -2;

	/* fill in the array ie[] for each internal edge
	 * (for the meaning of ie[], refer to the 'header.h' file) */ 
	for ( i = 0; i < es_size; i++ )
		for ( j = 0; j < 4; j++ ) {

			/* let pe be the neighbouring edge of es[i] given by e[j] */
			pe = es[i].e[j];

			/* test whether pe is an internal edge and whether the pair (i, pe) is still to be handled */
			if ( pe > i ) {
				
				/* find the index of the edge i in the array of neigbours of the edge es[pe] */
				k = 0;
				while ( es[pe].e[k] != i ) k++;

				/* fill in the indices */
				es[i].ie[j] = ( unsigned char ) k;
				es[pe].ie[k] = ( unsigned char ) j;
			}
		}
}
